# ProJor (Project Interpolator)

Projor allows you to generate code in your project, based on a model. Your model consists of _schema_, and _data_ files. This allows you to store any model alongside your projects. You can create _templates_, which operate on data to generate code. You can also extend the logic of the code generator, by providing your own _Handlebars helper functions_ in your project. You can also provide _queries_, which are virtual data collections, that are calculated from persistent data collections.

## Online resources

You can find further information about ProJor online:

* [ProJor Documentation](https://docs.projor.io)
* [ProJor License](https://license.projor.io)

For support, please contact SIOCODE at `support@siocode.hu`.

## Schema

Projor uses its own schema structure. Data can only be stored, so that it adheres to a schema. Each schema is a separate `.pschema.yaml` file. When referencing a schema, its `id` is used; all known schemas are loaded into memory, and schema references are resolved from this collection.

Example schema:

```yaml
id: company.com/app/Entity
name: Entity
description: A persistent entity
fields:
- type: string
  name: tableName
  description: The name of the table in the database
  required: true
```

### TypeScript schema

The following TypeScript files describe the schema structure:

```typescript
export type PSchemaFieldType_Basic =
    "string"
    | "integer"
    | "float"
    | "boolean";

export type PSchemaFieldType =
    PSchemaFieldType_Basic
    | "attachment"
    | "reference"
    | "enum";

export type PSchemaBaseField = {
    name: string;
    description: string;
    type: PSchemaFieldType;
    required?: boolean;
    multiple?: boolean;
}

export type PSchemaBasicField = PSchemaBaseField & {
    type: PSchemaFieldType_Basic;
    default?: string | number | boolean;
};

export type PSchemaReferenceOrAttachmentField = PSchemaBaseField & {
    type: "reference" | "attachment";
    references: string;
    _referencesSchema?: PSchema;
};

export type PSchemaEnumField = PSchemaBaseField & {
    type: "enum";
    enum: Array<string>;
};

export type PSchemaField =
    PSchemaBasicField
    | PSchemaReferenceOrAttachmentField
    | PSchemaEnumField;

export type PSchema = {
    id: string;
    name: string;
    description: string;
    fields: Array<PSchemaField>;
    _sourcePath?: string;
}
```

## Data

Data is stored in `.pdata.yaml` files. Each data file must adhere to a schema. The schema is referenced by the `schema` field. The `id` field is used to reference the data file. Each data file contains a list of objects, and the properties of these objects are defined by the schema.

* Every data item must have a `name`, and a `description` field. Name must be unique within the data collection.
* There are basic properties, such as `string`, `integer`, `float`, `boolean` and `enum` fields
* Attachment properties are used to embed data items in another data item. Attachments must also reference a schema.
* Reference properties are used to reference data items from another data collection. References always use a fully-qualified identifier, which is constructed like this: `{{data collection id}}#{{data item name}}`.
* Any field can be made a list by adding the `multiple` property.

Example data:

```yaml
id: company.com/app/epics
schema: company.com/app/Epic
objects:
  - name: PROJ-101
    description: Model user entities for the app
  - name: PROJ-192
    description: Model TODO entities for the app
  - name: PROJ-123
    description: Asdasd
```

### Global data

You may store _global data_ in `.pglobal.yaml` files. Global data has no schema. Data from these files are loaded, and merged into the `$` variable. You can reference global data in _any_ template, or filename template.

For example, you might create the `project.pglobal.yaml` file with the following content:

```yaml
scope: company
version: 1.2.3
```

Then, you can reference the `$.scope` and `$.version` variables in your templates.

### Template sharing

If you use _git submodules_ to share your templates accross multiple projects, it is possible to **override entire data collections** in your local project. You can use the `override: true` attribute in `.pdata.yaml` or `.pdata.json` files, so that the local collection overrides the shared collection (possibly loaded from the submodule).

### TypeScript data

The following TypeScript files describe the data structure:

```typescript
export type PObject = {
    name: string;
    description: string;
} & Record<string, any>;

export interface PData {
    id: string;
    objects: Array<PObject>;
    schema?: string;
    _resolvedSchema?: PSchema;
    _sourcePath?: string;
}
```

## Templates

You may use the following types of templates:

| Type | Loop | Description |
| -- | -- | -- |
| File template | For Each | Creates a file for each object in the data collection |
| File template | All Of | Creates a single file, using all objects in the data collection |
| File template | Map | Creates a single file, using a custom context from multiple data collections |
| Infix template | For Each | Generates code blocks for each object in the data collection, and inserts them into the target file, using the specified delimiter to join the blocks. |
| Infix template | All Of | Generates a single code block, using all objects in the data collection, and inserts it into the target file. |
| Infix template | Map | Generates a single code block, using a custom context from multiple data collections, and inserts it into the target file. |

Every template is a `.ptemplate.mustache` file in the `.projor` directory, and consists of a _JSON blocks_, and the _template content_. These two blocks are separated by a line containing only `---`.

In `forEach` and `allOf` clauses, you may use either a single data collection (string), or merge multiple data collections (array of strings). You may also either use a string or an array of strings for each entry in the `map` clause for the same effect.

You can also reference _schemas_ in `forEach`, `allOf` and `map` clauses. This comes handy, if you have multiple dynamic collections. ProJor will collect all data collections of a given schema, and merge them into a single collection, whose name is the schema id. Note, that you will need to use globally unique names for your data objects of the same schema.

You can also create so-called "_once templates_". This allows you to quickly generate new files, that you will modify later on. The code generator will not overwrite these files, if they already exist. The code generator will also not remove these files, if you remove the data objects that resulted in their generation.

You may also optionally specify the `formatUsing` value to apply formatting to the generated code. You can use the following values:

* `babel` (for JavaScript)
* `typescript` (for TypeScript)
* `java` (for Java)
* `graphql` (for GraphQL)
* `sql` (for SQL)
* `json` (for JSON)
* `yaml` (for YAML)
* `html` (for HTML)
* `css` (for CSS)
* `scss` (for SCSS)

In `forEach` templates, you may also use _conditional generation_. Place your condition between `@if` and `@endif` before the `---` separator in your template. The condition must be _valid JavaScript code_, that evaluates to a _function_, which takes a single argument, and returns a `boolean`.

```
@if (data) => data.name.startsWith("A") @endif
{ "forEach": "entities", "filename": "{{name}}.txt" }
---
{{name}} - {{description}}
```

Example for each file template:

```
{
    "forEach": "company.com/app/entities",
    "filename": "{{name}}.md"
}
---
# {{name}}

_{{description}}_
```

Example all of file template:

```
{
    "allOf": "company.com/app/entities",
    "filename": "ENTITIES.md"
}
---
| Name | Description |
| -- | -- |
{{#objects}}
| {{name}} | {{description}} |
{{/objects}}
```

Example map file template:

```
{
    "map": {
        "e": "entities",
        "s": "services"
    },
    "filename": "DOC.md"
}
---
# Entities

{{#each e}}
* {{name}}
{{/each}}

# Services

{{#each s}}
* {{name}}
{{/each}}
```

Example for each infix template:

```
{
    "forEach": "company.com/app/entities",
    "filename": "README.md",
    "infix": {
        "start": "| Entity Name | Entity Description |\n| --- | --- |\n",
        "end": "\n\nThis table describes entities.",
        "join": "\n"
    }
}
---
| {{name}} | {{description}} |
```

Example all of infix template:

```
{
    "allOf": "company.com/app/entities",
    "filename": "README.md",
    "infix": {
        "start": "# Entities\n",
        "end": "\n# REST API"
    }
}
---
| Name | Description |
| -- | -- |
{{#objects}}
| {{name}} | {{description}} |
{{/objects}}

This table describes entities.
```

_TODO: write the README.md for this template as well for clarity_

Example map infix template:

_TODO: write example_

### Partials

You may create _reusable sub-templates_, called _partials_. Partials are files in the `.projor` directory with the `.partial.mustache` extension. The name of the file before the extension will be the name of the partial.

Suppose you created the `.projor/myPartial.partial.mustache` file with the following content:

```
Partial content: {{name}}
```

Then, in your template you can invoke it as follows:

```
{
    "forEach": "my-data-collection",
    "filename": "{{name.txt}}"
}
---
Hi there!

{{> myPartial name=name }}
```

Given you have a single object like `{"name": "test"}` in your `my-data-collection` data collection, the above template will generate the following file:

Filename: `test.txt`

```
Hi there!

Partial content: test
```

### Template sharing

If you use _git submodules_ to share your templates accross multiple projects, it is possible to **override entire code templates** in your local project. You can use the `"override": true` attribute in the template metadata JSON block, so that the local template overrides the shared template (possibly loaded from the submodule).

### TypeScript templates

The following TypeScript files describe the template structure:

```typescript
export interface PTemplateInfix {
    start: string;
    end: string;
    join?: string;
}

export type BasePTemplate = {
    template: string;
    filename: string;
    infix?: PTemplateInfix;
    helpers?: Array<string>;
    formatUsing?: string;
    _directory: string;
    _templateContent?: string | null;
    _sourcePath?: string;
}

export type PTemplateWithForEach = {
    forEach: string | Array<string>;
} & BasePTemplate;
export type PTemplateWithAllOf = {
    allOf: string | Array<string>;
} & BasePTemplate;
export type PTemplateWithMap = {
    map: Record<string, string | Array<string>>;
} & BasePTemplate;

export type PTemplate = PTemplateWithForEach | PTemplateWithAllOf;
```

## Handlebars helpers

Helpers are JavaScript files in the `.projor` directory. The helper script must evaluate to a function, that returns an object. Each property in the object must be a function, these will be registered as helpers to Handlebars.

Example helper:

```javascript
(function helpers() {
    return {
        javaColumnType: function (typeName) {
            if (typeName === "ID") {
                return "Long";
            } else if (typeName === "STRING") {
                return "String";
            }
            return typeName;
        }
    }
})
```

Then this helper can be used as `{{javaColumnType type}}` in the template.

Helpers can be referenced in templates. You must add `file://` to the helper path, and provide relative path to the `.ptemplate.mustache` file itself. 

The following is a more complex example, that shows how a helper script from `.projor/helpers/entity-java.helpers.js` can be included from `.projor/entities.md.ptemplate.mustache`:

```
{
    "forEach": "company.com/app/entities",
    "helpers": [
        "file://./helpers/entity-java.helpers.js"
    ],
    "filename": "ENTITIES.md",
    "infix": {
        "start": "# Entities\n\n",
        "end": "\n\n# Summary",
        "join": "\n\n"
    }
}
---
## {{name}}

_Introduced in: {{epicLink.name}}_

{{description}}

### Columns

{{#columns}}
* `{{name}}`: **{{type}}** - _{{description}}_
{{/columns}}
```

### Built-in helpers

* `camelCase` - converts a string to camelCase
    * `{{camelCase "Hello World"}}` -> `helloWorld`
* `pascalCase` - converts a string to PascalCase
    * `{{pascalCase "Hello World"}}` -> `HelloWorld`
* `snakeCase` - converts a string to snake_case
    * `{{snakeCase "Hello World"}}` -> `hello_world`
* `kebabCase` - converts a string to kebab-case
    * `{{kebabCase "Hello World"}}` -> `hello-world`
* `constantCase` - converts a string to CONSTANT_CASE
    * `{{upperCase "Hello World"}}` -> `HELLO_WORLD`
* `dotCase` - converts a string to dot.case
    * `{{dotCase "Hello World"}}` -> `hello.world`
* `pathCase` - converts a string to path/case
    * `{{pathCase "Hello World"}}` -> `hello/world`
* `lowerCase` - converts a string to lower case
    * `{{lowerCase "Hello World"}}` -> `hello world`
* `upperCase` - converts a string to UPPER CASE
    * `{{upperCase "Hello World"}}` -> `HELLO WORLD`
* `boolAnd` - ands multiple boolean values
    * `{{#if (boolAnd flagOne flagTwo flagThree)}}Something{{/if}}`
* `boolOr` - ors multiple boolean values
    * `{{#if (boolOr flagOne flagTwo flagThree)}}Something{{/if}}`
* `boolNot` - negates a boolean value
    * `{{#if (boolNot flag)}}Something{{/if}}`
* `startsWith` - checks if a string starts with another string
    * `{{#if (startsWith string "Hello")}}Something{{/if}}`
* `endsWith` - checks if a string ends with another string
    * `{{#if (endsWith string "World")}}Something{{/if}}`
* `contains` - checks if a string contains another string
    * `{{#if (contains string "Hello")}}Something{{/if}}`
* `matches` - checks if a string matches a regular expression
    * `{{#if (matches string "Hello")}}Something{{/if}}`
* `equals` - checks if the two values are equal
    * `{{#if (equals valueOne valueTwo)}}Something{{/if}}`

## Queries

Queries can be used to create _virtual data collections_. Values are calculated from the _persistent data collections_. Then, these virtual data collections can also be interpolated in templates, by referencing their ids.

Creating a query is similar to creating a helper: you must define them in `.pquery.js` files inside the `.projor` directory. The query JavaScript file should contain code, that is going to be the body of a function. It must include a return statement, that returns an object, which adheres to the following TypeScript interface:

```typescript
export interface PQuery {
    id: string;
    schema: string;
    name?: string;
    description?: string;
    result: Array<PObject>;
}
```

The query code receives the `project: PProject` parameter, which is the resolved project before queries are executed. You may perform any arbitrary filtering, or projecting logic in your query.

The following is a complex example of a query:

```javascript
const entitiesData = project.data.find(
    d => d.id === "company.com/app/entities"
);
const epicsData = project.data.find(
    d => d.id === "company.com/app/epics"
);
const entities = entitiesData.objects;
const epics = epicsData.objects;

const queryResult = epics.map(
    epic => ({
        name: epic.name,
        description: epic.description,
        epic: `company.com/app/epics#${epic.name}`,
        entity: entities.filter(
            e => e.$epicLink === `company.com/app/epics#${epic.name}`
        ).map(
            e => `company.com/app/entities#${e.name}`
        )
    })
);

return {
    id: "company.com/app/epics-with-entities",
    result: queryResult,
    schema: "company.com/app/EpicWithEntity"
};
```
